import logging
import tracemalloc
from datetime import datetime
from tracemalloc import _format_size
from typing import Optional

from twisted.internet import reactor
from twisted.internet.threads import deferToThread

from peek_plugin_base.util.PeekPsUtil import PeekPsUtil

logger = logging.getLogger(__name__)


def rpad(val, count):
    val = str(val)
    return val + max(0, (count - len(val))) * " "


def lpad(val, count):
    val = str(val)
    return max(0, (count - len(val))) * " " + val


def center(val, width=80):
    return max(0, int((width - len(val)) / 2)) * " " + val


PEEK_MEM_DUMP_STACKTRACE = 2**0  # 1
PEEK_MEM_DUMP_VORTEX_OBSERVABLE_CACHE = 2**2  # 2
PEEK_MEM_DUMP_VORTEX_JSONABLE = 2**3  # 4
PEEK_MEM_DUMP_VORTEX_PUSH_PRODUCER = 2**4  # 8


def _formatTracemallocTraceback(top, size, count):
    sitePkgs = "site-packages"

    snapshot = tracemalloc.take_snapshot()

    # Get the objects by their line of malloc
    tracebackStats = list(
        filter(
            lambda s: s.size > size or s.count > count,
            snapshot.statistics("traceback"),
        )
    )[:top]

    text = "The traceback module is using %s memory\n" % _format_size(
        tracemalloc.get_tracemalloc_memory(), False
    )

    if not tracebackStats:
        text += "There are no large tracemalloc objects from traceback\n"
        return text

    text += "\n"

    # ---------------
    # Print the traceback summary

    for stat in tracebackStats:
        average = stat.size / stat.count

        tracePath = str(stat.traceback)
        if sitePkgs in tracePath:
            tracePath = tracePath[
                tracePath.index(sitePkgs) + len(sitePkgs) + 1 :
            ]

        text += "   size:%s" % lpad(_format_size(stat.size, False), 10)
        text += "   count:%s" % lpad(stat.count, 10)
        text += "   average:%s" % lpad(_format_size(average, False), 10)
        text += "   " + tracePath

        text += "\n"

        for line in stat.traceback.format():
            text += (" " * 8) + line + "\n"

        text += "\n"

    # ---------------
    # Print the lineno summary

    text += (
        " "
        + rpad("COUNT", 10)
        + " "
        + rpad("SIZE", 10)
        + " "
        + rpad("AVERAGE", 10)
        + " "
        + "MALLOC LINE"
        + "\n"
    )

    for stat in tracebackStats:
        average = stat.size / stat.count

        tracePath = str(stat.traceback)
        if sitePkgs in tracePath:
            tracePath = tracePath[
                tracePath.index(sitePkgs) + len(sitePkgs) + 1 :
            ]

        text += (
            " "
            + rpad(str(stat.count), 10)
            + " "
            + rpad(_format_size(stat.size, False), 10)
            + " "
            + rpad(_format_size(average, False), 10)
            + " "
            + tracePath
            + "\n"
        )

    text += "\n"

    return text


def _formatJsonableSummary(top, size):
    from vortex.Jsonable import Jsonable

    jsonableDump = Jsonable.memoryLoggingDump(top=top, over=size)

    if not jsonableDump:
        return "There are no large Vortex Jsonable Objects\n"

    text = " " + rpad("COUNT", 10) + " " + "OBJECT TYPE" + "\n"
    for objectType, count in jsonableDump:
        text += " " + rpad(str(count), 10) + " " + objectType + "\n"

    return text


def _formatObservableCacheSummary(top, size):
    from vortex.handler.TupleDataObservableCache import _CachedSubscribedData

    vortexCacheDump = _CachedSubscribedData.memoryLoggingDump(
        top=top, over=size
    )

    if not vortexCacheDump:
        return "There are no large Vortex Observable Caches\n"

    text = (
        " "
        + rpad("QUEUE ITEMS", 10)
        + " "
        + rpad("TOTAL SIZE", 10)
        + " "
        + "REMOTE VORTEX"
        + "\n"
    )

    for objectType, count, total in vortexCacheDump:
        text += (
            " "
            + rpad(str(count), 10)
            + " "
            + rpad(str(total), 10)
            + " "
            + objectType
            + "\n"
        )

    return text


def _formatVortexPushProducerSummary(top, msgs):
    from vortex.VortexWritePushProducer import VortexWritePushProducer

    stats = VortexWritePushProducer.memoryLoggingDump(top=top, msgs=msgs)

    if not stats:
        return "There are no large producer queues\n"

    text = (
        " "
        + rpad("MSGS", 10)
        + " "
        + rpad("TOTAL", 10)
        + " "
        + "REMOTE VORTEX"
        + "\n"
    )

    for objectType, count, total in stats:
        text += (
            " "
            + rpad(str(count), 10)
            + " "
            + rpad(str(total), 10)
            + " "
            + objectType
            + "\n"
        )

    return text


def dumpMemObjectToFile(
    debugMask: int = 0x11111111, serviceName: Optional[str] = None
):
    TRACEMALLOC_STACK_SIZE = 6

    # Start tracemalloc logging
    if debugMask & PEEK_MEM_DUMP_STACKTRACE:
        tracemalloc.start(TRACEMALLOC_STACK_SIZE)

    _dumpMemObjectToFile(debugMask, serviceName)


def _dumpMemObjectToFile(debugMask, serviceName):
    import os
    import pytz

    TRACEMALLOC_STACK_SIZE = 6

    class _State:
        lastClearDate = datetime.now(pytz.utc)

    TOP = 10
    COUNT_MIN = 10000
    TOTAL_SIZE = 1 * 1024 * 1024
    INDIVIDUAL_SIZE = 10 * 1024

    startTime = datetime.now(pytz.utc)
    # This is useful for debugging c binding memory leaks
    # roots = objgraph.get_leaking_objects()
    # objgraph.show_refs(roots[:3], refcounts=True, filename='/Users/peek/stash/roots.png')
    flushRequired = (
        3600 < (datetime.now(pytz.utc) - _State.lastClearDate).seconds
    )
    if flushRequired:
        _State.lastClearDate = datetime.now(pytz.utc)
    homeDir = os.path.expanduser("~/memdump-%s.log" % serviceName)
    with open(homeDir, "a") as f:
        # Write the start datetime
        f.write("=" * 80 + "\n")
        f.write("START - " + str(startTime) + "\n")

        if flushRequired:
            f.write("THE HOURLY FLUSH OCCURRED\n")

        # Write the total process memory
        f.write("-" * 80 + "\n")
        f.write(
            "Total python processes memory usage: "
            + rpad(_format_size(PeekPsUtil().memory_info.rss, False), 10)
            + "\n"
        )

        if debugMask & PEEK_MEM_DUMP_STACKTRACE:
            if flushRequired:
                tracemalloc.stop()
                tracemalloc.start(TRACEMALLOC_STACK_SIZE)

            f.write("-" * 80 + "\n")
            f.write(center("Python Tracemalloc Information") + "\n")
            f.write(_formatTracemallocTraceback(TOP, TOTAL_SIZE, COUNT_MIN))

        if debugMask & PEEK_MEM_DUMP_VORTEX_JSONABLE:
            # Write the _loop of the Jsonable objects
            f.write("-" * 80 + "\n")
            f.write(center("Vortex Jsonable Objects") + "\n")
            f.write(_formatJsonableSummary(TOP, INDIVIDUAL_SIZE))

        if debugMask & PEEK_MEM_DUMP_VORTEX_OBSERVABLE_CACHE:
            # Write the _loop of the cached vortex payloads
            f.write("-" * 80 + "\n")
            f.write(center("Vortex Observable Caches") + "\n")
            f.write(_formatObservableCacheSummary(TOP, INDIVIDUAL_SIZE))

        if debugMask & PEEK_MEM_DUMP_VORTEX_PUSH_PRODUCER:
            # Write the _loop of the cached vortex payloads
            f.write("-" * 80 + "\n")
            f.write(center("Vortex Write Push Producer") + "\n")
            f.write(_formatVortexPushProducerSummary(10, 1))

        # Write the end date
        f.write("-" * 80 + "\n")
        f.write(
            "END - Time taken : %s seconds"
            % str(datetime.now(pytz.utc) - startTime)
            + "\n"
        )
        f.write("-" * 80 + "\n")


def setupMemoryDebugging(serviceName: Optional[str] = None, debugMask: int = 0):
    import pytz

    TRACEMALLOC_STACK_SIZE = 6

    class _State:
        lastClearDate = datetime.now(pytz.utc)

    # Start tracemalloc logging
    if debugMask & PEEK_MEM_DUMP_STACKTRACE:
        tracemalloc.start(TRACEMALLOC_STACK_SIZE)

    # Start JSonable logging
    from vortex.Jsonable import Jsonable

    if debugMask & PEEK_MEM_DUMP_VORTEX_JSONABLE:
        Jsonable.setupMemoryLogging()

    from vortex.handler.TupleDataObservableCache import _CachedSubscribedData

    if debugMask & PEEK_MEM_DUMP_VORTEX_OBSERVABLE_CACHE:
        # Start cached encoded payload logging
        _CachedSubscribedData.setupMemoryLogging()

    from vortex.VortexWritePushProducer import VortexWritePushProducer

    if debugMask & PEEK_MEM_DUMP_VORTEX_PUSH_PRODUCER:
        # Start cached encoded payload logging
        VortexWritePushProducer.setupMemoryLogging()

    logger.warning("Memory Logging is enabled.")

    def _loop():
        dumpMemObjectToFile(debugMask, serviceName)

        reactor.callLater(60, deferToThread, _loop)

    _loop()



import json
import logging
import os
from base64 import b64decode
from base64 import b64encode

from twisted.internet import protocol
from twisted.internet import task
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.VortexFactory import VortexFactory

from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_child_state_protocol import (
    PluginSubprocChildStateProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    LOGGING_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_vortex_msg_tuple import (
    PluginSubprocVortexMsgTuple,
)


logger = logging.getLogger(__name__)


class PluginSubprocParentProtocol(protocol.ProcessProtocol):
    VORTEX_UUID_UPDATE_PERIOD = 30

    def __init__(self, subprocessGroupName):
        self._dataBytesArray = bytearray()
        self._logData = b""
        self._pluginStateData = b""

        self._vortexUpdateLoopingCall = task.LoopingCall(
            self._sendUpdatedVortexUuids
        )

        self._lastPluginStateCommandDeferred = None

        self._loggerForChild = logging.getLogger(
            f"subproc:{subprocessGroupName}"
        )

    @inlineCallbacks
    def childDataReceived(self, childFD: int, data: bytes):
        if childFD == VORTEX_MSG_FROM_CHILD_FD:
            yield self.outReceived(data)
        elif childFD == LOGGING_FROM_CHILD_FD:
            yield self.errReceived(data)
        # elif childFD == VORTEX_UUID_FROM_CHILD_FD:
        #     not used
        elif childFD == PLUGIN_STATE_FROM_CHILD_FD:
            yield self._pluginStateDataReceived(data)
        else:
            raise NotImplementedError(
                f"We didn't expect to get data from {childFD}"
            )

    def connectionMade(self):
        self._vortexUpdateLoopingCall.start(self.VORTEX_UUID_UPDATE_PERIOD)

    def inConnectionLost(self):
        """
        This will be called when stdin is closed.
        """
        if self._vortexUpdateLoopingCall:
            self._vortexUpdateLoopingCall.stop()
            self._vortexUpdateLoopingCall = None

    # ---------------------------------
    # Handle sending vortex uuid updates to subprocess

    @inlineCallbacks
    def _sendUpdatedVortexUuids(self):
        encodedUuids = yield self._encodeVortexUuids()
        self.transport.writeToChild(VORTEX_UUID_TO_CHILD_FD, encodedUuids)
        self.transport.writeToChild(VORTEX_UUID_TO_CHILD_FD, b".")

    @deferToThreadWrapWithLogger(logger)
    def _encodeVortexUuids(self):
        remoteInfoTuples = [
            [info.name, info.uuid]
            for info in VortexFactory.getRemoteClientVortexInfos()
        ]
        encodedUuids = b64encode(json.dumps(remoteInfoTuples).encode())
        return encodedUuids

    # ---------------------------------
    # Handle sending vortex messages from clients

    def errReceived(self, data: bytes):
        self._logData += data

        while b"\n" in self._logData:
            message, self._logData = self._logData.split(b"\n", 1)
            if not message:
                continue

            message = message.decode()

            if ":" not in message:
                self._loggerForChild.error(message)

            else:
                severity, logMsg = message.split(":", 1)
                if severity == "DEBUG":
                    self._loggerForChild.debug(logMsg)
                elif severity == "INFO":
                    self._loggerForChild.info(logMsg)
                elif severity == "WARNING":
                    self._loggerForChild.warning(logMsg)
                elif severity == "ERROR":
                    self._loggerForChild.error(logMsg)
                else:
                    self._loggerForChild.error(message)

    @inlineCallbacks
    def outReceived(self, data):
        self._dataBytesArray.append(data)

        if b"." not in data:
            return

        # Avoid accidentally referencing this
        del data

        joinedData = bytes(self._dataBytesArray)

        while b"." in joinedData:
            message, joinedData = joinedData.split(b".", 1)
            if not message:
                continue

            yield self._sendVortexMsgFromChild(message)

        self._dataArray = bytearray(joinedData)

    @inlineCallbacks
    def _sendVortexMsgFromChild(self, message: bytes):
        vortexMsgTuple = yield self._decodeVortexMsgTuple(message)

        yield VortexFactory.sendVortexMsg(
            vortexMsgs=vortexMsgTuple.vortexMsgs,
            destVortexUuid=vortexMsgTuple.vortexUuid,
        )

    @deferToThreadWrapWithLogger(logger)
    def _decodeVortexMsgTuple(self, message: bytes):
        return PluginSubprocVortexMsgTuple().fromJsonDict(
            json.loads(b64decode(message).decode())
        )

    # ---------------------------------
    # Handle the plugin state changes

    def sendPluginLoad(self, pluginName: str) -> Deferred:
        return self._sendPluginStateCommand(
            pluginName, PluginSubprocChildStateProtocol.COMMAND_LOAD
        )

    def sendPluginStart(self, pluginName: str) -> Deferred:
        return self._sendPluginStateCommand(
            pluginName, PluginSubprocChildStateProtocol.COMMAND_START
        )

    def sendPluginStop(self, pluginName: str) -> Deferred:
        return self._sendPluginStateCommand(
            pluginName, PluginSubprocChildStateProtocol.COMMAND_STOP
        )

    def sendPluginUnload(self, pluginName: str) -> Deferred:
        return self._sendPluginStateCommand(
            pluginName, PluginSubprocChildStateProtocol.COMMAND_UNLOAD
        )

    @inlineCallbacks
    def _sendPluginStateCommand(self, pluginName: str, command: bytes):
        assert not self._lastPluginStateCommandDeferred, (
            "There is already a pending " "command"
        )
        self._lastPluginStateCommandDeferred = Deferred()

        self.transport.writeToChild(
            PLUGIN_STATE_TO_CHILD_FD, f"{pluginName}:{command}".encode()
        )
        self.transport.writeToChild(PLUGIN_STATE_TO_CHILD_FD, b"\n")

        yield self._lastPluginStateCommandDeferred
        self._lastPluginStateCommandDeferred = None

    def _pluginStateDataReceived(self, data):
        self._pluginStateData += data

        while b"\n" in self._pluginStateData:
            message, self._pluginStateData = self._pluginStateData.split(
                b"\n", 1
            )
            if not message:
                continue

            if not self._lastPluginStateCommandDeferred:
                error = (
                    f"We got a result when we have no deferred to "
                    f"call, result={message}"
                )
                logger.error(error)
                raise Exception(error)

            pluginName, result = message.decode().split(":", 1)

            if result == PluginSubprocChildStateProtocol.COMMAND_SUCCESS:
                self._lastPluginStateCommandDeferred.callback(result)

            else:
                self._lastPluginStateCommandDeferred.errback(result)

            self._lastPluginStateCommandDeferred = None



import logging

from jsoncfg.value_mappers import require_string, require_dict

logger = logging.getLogger(__name__)


class PeekFileConfigSqlAlchemyMixin:
    @property
    def dbConnectString(self):
        default = "postgresql+psycopg://peek:PASSWORD@127.0.0.1/peek"
        with self._cfg as c:
            return c.sqlalchemy.connectUrl(default, require_string)

    @property
    def dbEngineArgs(self):
        default = {
            "echo": False,  # Print every SQL statement executed
            "pool_size": 20,  # Number of connections to keep open
            "max_overflow": 50,
            # Number that the pool size can exceed when required
            "pool_timeout": 60,  # Timeout for getting conn from pool
            "pool_recycle": 600,  # Reconnect?? after 10 minutes
            # This supersedes 'use_batch_mode': True,
        }
        with self._cfg as c:
            val = c.sqlalchemy.engineArgs(default, require_dict)
            # Upgrade depreciated psycopg setting.
            if val.get("use_batch_mode") == True:
                del val["use_batch_mode"]
                c.sqlalchemy.engineArgs = val

            if "client_encoding" not in val:
                val["client_encoding"] = "utf8"
                c.sqlalchemy.engineArgs = val

            if "executemany_mode" in val:
                del val["executemany_mode"]

            return val



from . import WindowsPatch

__version__ = '4.1.2'


class PeekPlatformConfig:
    """Peek Platform Config

    This is really a GLOBAL pettern, It should be replaced at some stage.
    (Maybe named as a factory?)

    This class is populated with data when the peek processes start.
    This is required so that peek_platform common code can access the other parts
    of the system, which are peek_agent_service, peek_logic_service, peek_worker_service.

    """

    # The component name of this part of the platform
    # EG, peek_logic_service, peek_worker_service, peek_agent_service
    componentName = None

    # The config accessor class
    config = None

    # The inherited class of PluginSwInstallManagerABC
    pluginSwInstallManager = None

    # The inherited class of PeekSwInstallManagerABC
    peekSwInstallManager = None

    # The instance of the PluginLoaderABC
    pluginLoader = None

    # Is this a subprocess instance of a service.
    isPluginSubprocess = False



import gzip
import logging
import os
import sys
from logging.handlers import SysLogHandler
from logging.handlers import TimedRotatingFileHandler

from pathlib import Path
from typing import Optional

import warnings
from cryptography.utils import CryptographyDeprecationWarning

warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)

LOG_FORMAT = "%(asctime)s %(levelname)s %(name)s:%(message)s"
DATE_FORMAT = "%d-%b-%Y %H:%M:%S"

logger = logging.getLogger(__name__)


def setupPeekLogger(serviceName: Optional[str] = None):
    logging.basicConfig(
        stream=sys.stdout,
        format=LOG_FORMAT,
        datefmt=DATE_FORMAT,
        level=logging.DEBUG,
    )
    logging.getLogger("ampq").setLevel(logging.INFO)
    logging.getLogger("celery.utils.functional").setLevel(logging.INFO)

    if serviceName:
        updatePeekLoggerHandlers(serviceName)


def _namer(name):
    return name + ".gz"


def _rotator(source, dest):
    READ_CHUNK = 512 * 1024
    if not os.path.exists(source):
        return

    with open(source, "rb") as sf:
        with gzip.open(dest, "wb") as f:
            data = sf.read(READ_CHUNK)
            while data:
                f.write(data)
                data = sf.read(READ_CHUNK)
    os.remove(source)


def updatePeekLoggerHandlers(
    serviceName: Optional[str] = None, daysToKeep=28, logToStdout=True
):
    rootLogger = logging.getLogger()
    logFormatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

    for handler in list(rootLogger.handlers):
        if isinstance(handler, TimedRotatingFileHandler):
            # Setup the file logging output
            rootLogger.removeHandler(handler)

        elif not sys.stdout.isatty() and not logToStdout:
            # Remove the stdout handler
            logger.info(
                "Logging to stdout disabled, see 'logToStdout' in config.json"
            )
            rootLogger.removeHandler(handler)

    fileName = str(Path.home() / ("%s.log" % serviceName))

    fh = TimedRotatingFileHandler(
        fileName, when="midnight", backupCount=daysToKeep
    )
    fh.setFormatter(logFormatter)
    fh.rotator = _rotator
    fh.namer = _namer
    rootLogger.addHandler(fh)


def setupLoggingToSyslogServer(host: str, port: int, facility: str):
    rootLogger = logging.getLogger()
    logFormatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

    logging.getLogger().addHandler(logging.StreamHandler())

    if facility not in SysLogHandler.facility_names:
        logger.info(list(SysLogHandler.facility_names))
        raise Exception("Syslog facility name is a valid facility")

    facilityNum = SysLogHandler.facility_names[facility]

    fh = SysLogHandler(address=(host, port), facility=facilityNum)
    fh.setFormatter(logFormatter)
    rootLogger.addHandler(fh)



import logging
import multiprocessing
import typing
from atexit import register
from logging.handlers import QueueHandler
from logging.handlers import QueueListener
from logging.handlers import TimedRotatingFileHandler

import celery
from celery import signals as celery_signals
from kombu import serialization
from peek_platform.file_config.PeekFileConfigWorkerMixin import (
    PeekFileConfigWorkerMixin,
)
from vortex.DeferUtil import noMainThread
from vortex.Payload import Payload

logger = logging.getLogger(__name__)


def vortexDumps(arg: typing.Tuple) -> str:
    noMainThread()
    # startTime = datetime.now(pytz.utc)
    try:
        return Payload(tuples=[arg])._toJson()
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexDumps took %s", (datetime.now(pytz.utc) - startTime))


def vortexLoads(jsonStr: str) -> typing.Tuple:
    noMainThread()
    # startTime = datetime.now(pytz.utc)
    try:
        return Payload()._fromJson(jsonStr).tuples[0]
    except Exception as e:
        logger.exception(e)
        raise
    # finally:
    #     logger.debug("vortexLoads took %s", (datetime.now(pytz.utc) - startTime))


serialization.register(
    "vortex",
    vortexDumps,
    vortexLoads,
    content_type="application/x-vortex",
    content_encoding="utf-8",
)


# -----------------------------------------------------------------------------


class BackendMixin:
    def exception_to_python(self, exc):
        """Convert serialized exception to Python exception."""
        import sys
        from kombu.utils.encoding import from_utf8
        from celery.utils.serialization import (
            create_exception_cls,
            get_pickled_exception,
        )

        EXCEPTION_ABLE_CODECS = frozenset({"pickle"})

        if not exc:
            return exc

        if not isinstance(exc, BaseException):
            exc_module = exc.get("exc_module")
            if exc_module is None:
                cls = create_exception_cls(from_utf8(exc["exc_type"]), __name__)
            else:
                exc_module = from_utf8(exc_module)
                exc_type = from_utf8(exc["exc_type"])
                try:
                    cls = getattr(sys.modules[exc_module], exc_type)
                except KeyError:
                    cls = create_exception_cls(
                        exc_type, celery.exceptions.__name__
                    )
            exc_msg = exc["exc_message"]
            args = exc_msg if isinstance(exc_msg, tuple) else [exc_msg]

            ### BEGIN CODE ADDED TO PATCH METHOD
            try:
                exc = cls(*args)

            except TypeError:
                exc = Exception("%s\n%s" % (cls, args))
            ### END CODE ADDED TO PATCH METHOD

            if self.serializer in EXCEPTION_ABLE_CODECS:
                exc = get_pickled_exception(exc)

        return exc


from celery.backends.base import Backend

Backend.exception_to_python = BackendMixin.exception_to_python


# -----------------------------------------------------------------------------


class ResultConsumerMixin:
    def cancel_for(self, task_id):
        import redis

        if not self._pubsub:
            return

        key = self._get_key_for_task(task_id)
        self.subscribed_to.discard(key)

        try:
            self._pubsub.unsubscribe(key)

        except redis.exceptions.ConnectionError:
            # redis.exceptions.ConnectionError: Error 32 while writing to socket. Broken pipe.
            logger.debug("Cancel_for failed due to redis error")

        except RuntimeError:
            # builtins.RuntimeError: dictionary changed size during iteration
            logger.debug("Cancel_for failed due to redis error")

        except AttributeError:
            # builtins.AttributeError: 'NoneType' object has no attribute 'sendall'
            logger.debug("Cancel_for failed due to redis error")


from celery.backends.redis import ResultConsumer

ResultConsumer.cancel_for = ResultConsumerMixin.cancel_for


# -----------------------------------------------------------------------------


def configureCeleryApp(
    app, workerConfig: PeekFileConfigWorkerMixin, forCaller: bool = False
):
    # Optional configuration, see the application user guide.
    app.conf.update(
        # On peek_logic_service, the thread limit is set to 10, these should be configurable.
        # broker_pool_limit=20,
        # Set the broker and backend URLs
        broker_url=workerConfig.celeryBrokerUrl,
        result_backend=workerConfig.celeryResultUrl,
        # Leave the logging to us
        worker_hijack_root_logger=False,
        # The time results will stay in redis before expiring.
        # I believe they are cleared when the results are obtained
        # from txcelery._DeferredTask
        # I assume the timer only starts once the task has finished.
        result_expires=60,
        task_serializer="vortex",
        # accept_content=['vortex'],  # Ignore other content
        accept_content=["pickle", "json", "msgpack", "yaml", "vortex"],
        result_serializer="vortex",
        enable_utc=True,
        # Default time in seconds before a retry of the task should be executed.
        # 3 minutes by default.
        default_retry_delay=2,
        broker_connection_retry_on_startup=True,
        # The maximum number of times to retry a task
        max_retries=5,
    )

    # Configure these only for the worker, this keeps the servers json clean.
    if not forCaller:
        # Optional configuration, see the application user guide.
        app.conf.update(
            # The number of tasks each worker will prefetch.
            worker_prefetch_multiplier=workerConfig.celeryTaskPrefetch,
            # The number of workers to have at one time
            worker_concurrency=workerConfig.celeryWorkerCount,
        )

        if workerConfig.celeryReplaceWorkerAfterTaskCount:
            app.conf.update(
                # The number of tasks a worker will process before it's replaced
                worker_max_tasks_per_child=workerConfig.celeryReplaceWorkerAfterTaskCount
            )

        if workerConfig.celeryReplaceWorkerAfterMemUsage:
            app.conf.update(
                # If a worker uses more than this amount of memory, it will be replaced
                # after the task completes.
                worker_max_memory_per_child=workerConfig.celeryReplaceWorkerAfterMemUsage
            )


from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC
from peek_platform.file_config.PeekFileConfigPlatformMixin import (
    PeekFileConfigPlatformMixin,
)


class _WorkerTaskConfigMixin(PeekFileConfigABC, PeekFileConfigPlatformMixin):
    pass


class QueuedTimedRotatingFileHandler:
    instance = None

    def __init__(self, handler: TimedRotatingFileHandler):
        self.queue = multiprocessing.Queue(-1)
        self.listener = QueueListener(self.queue, handler)
        self.listener.start()
        register(self.stop)

    def stop(self):
        self.listener.stop()


@celery_signals.after_setup_logger.connect
def configureCeleryLogging(*args, **kwargs):
    from peek_plugin_base.PeekVortexUtil import peekWorkerName

    # Fix the loading problems windows has
    # from peek_platform.util.LogUtil import setupPeekLogger
    # setupPeekLogger(peekWorkerName)

    from peek_platform import PeekPlatformConfig

    PeekPlatformConfig.componentName = peekWorkerName
    config = _WorkerTaskConfigMixin()

    # Set default logging level
    logging.root.setLevel(config.loggingLevel)

    if config.loggingLevel != "DEBUG":
        for name in (
            "celery.worker.strategy",
            "celery.app.trace",
            "celery.worker.job",
        ):
            logging.getLogger(name).setLevel(logging.WARNING)

    # replace TimedRotatingFileHandler with multiprocess supported version
    logger = kwargs.get("logger", None)
    if logger is None:
        return

    for handler in logger.handlers:
        if not isinstance(handler, TimedRotatingFileHandler):
            continue

        if QueuedTimedRotatingFileHandler.instance is not None:
            continue

        QueuedTimedRotatingFileHandler.instance = (
            QueuedTimedRotatingFileHandler(handler)
        )

        updatedHandler = QueueHandler(
            QueuedTimedRotatingFileHandler.instance.queue
        )

        logger.removeHandler(handler)
        logger.addHandler(updatedHandler)



from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class PluginSubprocVortexPayloadEnvelopeTuple(Tuple):
    __tablename__ = "PluginSubprocVortexPayloadEnvelopeTuple"
    __tupleType__ = "peek_platform." + __tablename__

    # Subprocess fields
    # The uuid of the vortex this message was received from
    vortexUuid: str = TupleField()
    vortexName: str = TupleField()

    # Call arguments
    payloadEnvelope: PayloadEnvelope = TupleField()



import gc
import logging
import os
import sys
from abc import ABCMeta
from abc import abstractmethod
from abc import abstractproperty
from collections import defaultdict
from importlib.util import find_spec
from typing import Optional
from typing import Tuple
from typing import Type

from jsoncfg.value_mappers import require_array
from jsoncfg.value_mappers import require_string
from twisted.internet.defer import maybeDeferred
from vortex.DeferUtil import vortexLogFailure

from peek_platform import PeekPlatformConfig
from twisted.internet.defer import inlineCallbacks

from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_platform_config_tuple import (
    PluginSubprocPlatformConfigTuple,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_vortex_msg_tuple import (
    PluginSubprocVortexMsgTuple,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_vortex_payload_envelope_tuple import (
    PluginSubprocVortexPayloadEnvelopeTuple,
)
from peek_plugin_base.PluginCommonEntryHookABC import PluginCommonEntryHookABC
from peek_plugin_base.PluginPackageFileConfig import PluginPackageFileConfig
from vortex.PayloadIO import PayloadIO
from vortex.Tuple import registeredTupleNames
from vortex.Tuple import removeTuplesForTupleNames
from vortex.Tuple import tupleForTupleName
from vortex.TupleAction import TupleGenericAction
from vortex.TupleAction import TupleUpdateAction
from vortex.TupleSelector import TupleSelector
from vortex.VortexUtil import _DebounceArgsTuple
from vortex.rpc.RPC import _VortexRPCArgTuple
from vortex.rpc.RPC import _VortexRPCResultTuple

from peek_plugin_base.simple_subproc.simple_subproc_task_call_tuple import (
    SimpleSubprocTaskCallTuple,
)
from peek_plugin_base.simple_subproc.simple_subproc_task_constructor_tuple import (
    SimpleSubprocTaskConstructorTuple,
)
from peek_plugin_base.simple_subproc.simple_subproc_task_result_tuple import (
    SimpleSubprocTaskResultTuple,
)

logger = logging.getLogger(__name__)

# This doesn't do anything, but it makes sure it's imported before any plugins import it.
TupleSelector()
TupleUpdateAction()
TupleGenericAction()
_VortexRPCResultTuple()
_VortexRPCArgTuple()
_DebounceArgsTuple()
SimpleSubprocTaskResultTuple()
SimpleSubprocTaskCallTuple()
SimpleSubprocTaskConstructorTuple()
PluginSubprocPlatformConfigTuple()
PluginSubprocVortexMsgTuple()
PluginSubprocVortexPayloadEnvelopeTuple()

corePlugins = [
    "peek_core_email",
    "peek_core_device",
    "peek_core_search",
    "peek_core_user",
    "peek_core_docdb",
    "peek_core_screen",
]


class PluginLoaderABC(metaclass=ABCMeta):
    _instance = None

    def __new__(cls, *args, **kwargs):
        assert (
            cls._instance is None
        ), "PluginServerLoader is a singleton, don't construct it"
        cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        self._loadedPlugins = {}
        self._loadedSubprocessGroups = {}

        self._vortexEndpointInstancesByPluginName = defaultdict(list)
        self._vortexTupleNamesByPluginName = defaultdict(list)

    @abstractproperty
    def _entryHookFuncName(self) -> str:
        """Entry Hook Func Name.
        Protected property
        :return: EG  "peekLogicEntryHook"

        """

    @abstractproperty
    def _entryHookClassType(self):
        """Entry Hook Class Type
        Protected property
        :return: EG  PluginLogicEntryHookABC

        """

    @abstractproperty
    def _platformServiceNames(self) -> [str]:
        """Platform Service Name
        Protected property
        :return: one or more of "logic", "worker", "agent", "client", "storage"

        """

    def pluginEntryHook(self, pluginName) -> Optional[PluginCommonEntryHookABC]:
        """Plugin Entry Hook

        Returns the loaded plugin entry hook for the plugin name.

        :param pluginName: The name of the plugin to load

        :return: An instance of the plugin entry hook

        """
        return self._loadedPlugins.get(pluginName)

    @inlineCallbacks
    def loadPlugin(self, pluginName):
        # Until we implement dynamic loading and unloading of plugins
        # make sure we don't load them twice, because we don't yet
        # support it.

        from peek_platform import PeekPlatformConfig

        if pluginName in self._loadedPlugins:
            raise Exception(
                "Plugin %s is already loaded, check config.json" % pluginName
            )

        try:
            self.unloadPlugin(pluginName)

            # Make note of the initial registrations for this plugin
            endpointInstancesBefore = set(PayloadIO().endpoints)
            tupleNamesBefore = set(registeredTupleNames())

            modSpec = find_spec(pluginName)
            if not modSpec:
                raise Exception(
                    "Failed to find package %s,"
                    " is the python package installed?" % pluginName
                )

            PluginPackage = modSpec.loader.load_module()

            pluginRootDir = os.path.dirname(PluginPackage.__file__)

            # Load up the plugin package info
            pluginPackageJson = PluginPackageFileConfig(pluginRootDir)
            pluginVersion = pluginPackageJson.config.plugin.version(
                require_string
            )
            pluginRequiresService = pluginPackageJson.config.requiresServices(
                require_array
            )

            # Make sure the service is required
            # Storage and Server are loaded at the same time, hence the intersection
            if not set(pluginRequiresService) & set(self._platformServiceNames):
                logger.debug(
                    "%s does not require %s, Skipping load",
                    pluginName,
                    self._platformServiceNames,
                )
                return

            configForService = pluginPackageJson.configForService(
                PeekPlatformConfig.componentName
            )

            # Get the entry hook class from the package
            entryHookGetter = getattr(
                PluginPackage, str(self._entryHookFuncName)
            )

            subprocessGroup = configForService.subprocessGroup(None)
            runInSubprocess = (
                subprocessGroup and not PeekPlatformConfig.isPluginSubprocess
            )

            RealEntryHookClass = entryHookGetter() if entryHookGetter else None

            if not RealEntryHookClass:
                logger.warning(
                    "Skipping load for %s, %s.%s is missing or returned None",
                    pluginName,
                    pluginName,
                    self._entryHookFuncName,
                )
                return

            if not issubclass(RealEntryHookClass, self._entryHookClassType):
                raise Exception(
                    "%s load error, Excpected %s, received %s"
                    % (pluginName, self._entryHookClassType, RealEntryHookClass)
                )

            # Are we going to run this plugin in a subprocess?
            if False:  # runInSubprocess:
                pass
                """
                from peek_platform.subproc_plugin_init.plugin_subproc_parent_main import (
                    PluginSubprocParentMain,
                )

                # We can run multiple plugins in one group
                # Have we already created this group?
                if subprocessGroup in self._loadedSubprocessGroups:
                    subprocParentMain = self._loadedSubprocessGroups[
                        subprocessGroup
                    ]
                else:
                    # Else, create it.
                    subprocParentMain = PluginSubprocParentMain(subprocessGroup)
                    self._loadedSubprocessGroups[
                        subprocessGroup
                    ] = subprocParentMain

                # Return the subprocParentMain, it has a __call__ method
                # to simulate the plugin class constructor
                EntryHookClass = subprocParentMain
                """

            else:
                EntryHookClass = RealEntryHookClass

            ### Perform the loading of the plugin
            yield self._loadPluginThrows(
                pluginName,
                EntryHookClass,
                pluginRootDir,
                tuple(pluginRequiresService),
            )

            yield self._setupStaticWebResourcesForPlugin(
                RealEntryHookClass, pluginName
            )

            # Make sure the version we have recorded is correct
            # JJC Disabled, this is just spamming the config file at the moment
            # PeekPlatformConfig.config.setPluginVersion(pluginName, pluginVersion)

            # Make note of the final registrations for this plugin
            self._vortexEndpointInstancesByPluginName[pluginName] = list(
                set(PayloadIO().endpoints) - endpointInstancesBefore
            )

            self._vortexTupleNamesByPluginName[pluginName] = list(
                set(registeredTupleNames()) - tupleNamesBefore
            )

            self.sanityCheckServerPlugin(pluginName)

        except Exception as e:
            logger.error("Failed to load plugin %s", pluginName)
            logger.exception(e)

    @abstractmethod
    def _loadPluginThrows(
        self,
        pluginName: str,
        EntryHookClass: Type[PluginCommonEntryHookABC],
        pluginRootDir: str,
        requiresService: Tuple[str, ...],
    ) -> None:
        """Load Plugin (May throw Exception)

        This method is called to perform the load of the module.

        :param pluginName: The name of the Peek App, eg "plugin_noop"
        :param EntryHookClass: The plugin entry hook class to construct.
        :param pluginRootDir: The directory of the plugin package,
         EG dirname(plugin_noop.__file__)

        """

    @inlineCallbacks
    def _setupStaticWebResourcesForPlugin(
        self, RealEntryHookClass, pluginName: str
    ):
        """Setup Static Web Resources For Plugin

        If this is running in a subprocess, we need to call this method on the
        actual plugin, not PluginSubprocParentMain
        """
        loadedPlugin = self._loadedPlugins[pluginName]

        yield RealEntryHookClass.setupStaticWebResources(loadedPlugin._platform)

    @inlineCallbacks
    def unloadPlugin(self, pluginName: str):
        oldLoadedPlugin = self._loadedPlugins.get(pluginName)

        if not oldLoadedPlugin:
            return

        del oldLoadedPlugin

        # Remove the registered endpoints
        for endpoint in self._vortexEndpointInstancesByPluginName[pluginName]:
            PayloadIO().remove(endpoint)
        del self._vortexEndpointInstancesByPluginName[pluginName]

        # Remove the registered tuples
        removeTuplesForTupleNames(
            self._vortexTupleNamesByPluginName[pluginName]
        )
        del self._vortexTupleNamesByPluginName[pluginName]

        yield self._unloadPluginPackage(pluginName)

    def listPlugins(self):
        def pluginTest(name):
            if not name.startswith("plugin_"):
                return False
            return os.path.isdir(os.path.join(self._pluginPath, name))

        plugins = os.listdir(self._pluginPath)
        plugins = list(filter(pluginTest, plugins))
        return plugins

    # ---------------
    # Core Plugins

    @inlineCallbacks
    def loadCorePlugins(self):
        for pluginName in corePlugins:
            yield self.loadPlugin(pluginName)

    @inlineCallbacks
    def startCorePlugins(self):
        # Start the Plugin
        for pluginName in corePlugins:
            if pluginName not in self._loadedPlugins:
                continue

            yield self._tryStart(pluginName)

    @inlineCallbacks
    def stopCorePlugins(self):
        # Start the Plugin
        for pluginName in corePlugins:
            if pluginName not in self._loadedPlugins:
                continue

            yield self._tryStop(pluginName)

    def unloadCorePlugins(self):
        for pluginName in corePlugins:
            if pluginName in self._loadedPlugins:
                self.unloadPlugin(pluginName)

    # ---------------
    # Optional Plugins

    @inlineCallbacks
    def loadOptionalPlugins(self):
        for pluginName in PeekPlatformConfig.config.pluginsEnabled:
            if pluginName.startswith("peek_core"):
                raise Exception("Core plugins can not be configured")
            yield self.loadPlugin(pluginName)

    @inlineCallbacks
    def startOptionalPlugins(self):
        # Start the Plugin
        for pluginName in PeekPlatformConfig.config.pluginsEnabled:
            if pluginName not in self._loadedPlugins:
                continue

            yield self._tryStart(pluginName)

    @inlineCallbacks
    def stopOptionalPlugins(self):
        # Start the Plugin
        for pluginName in reversed(PeekPlatformConfig.config.pluginsEnabled):
            if pluginName not in self._loadedPlugins:
                continue

            yield self._tryStop(pluginName)

    def unloadOptionalPlugins(self):
        for pluginName in reversed(PeekPlatformConfig.config.pluginsEnabled):
            if pluginName in self._loadedPlugins:
                self.unloadPlugin(pluginName)

        remainingOptionalPlugins = list(
            filter(lambda n: not n.startswith("peek_core"), self._loadedPlugins)
        )

        if remainingOptionalPlugins:
            logger.debug(remainingOptionalPlugins)
            raise Exception("Some plugins are still loaded")

    # ---------------
    # Standalone Plugins

    @inlineCallbacks
    def loadStandalonePlugin(self, pluginName: str):
        yield self.loadPlugin(pluginName)

    @inlineCallbacks
    def startStandalonePlugin(self, pluginName: str):
        yield self._tryStart(pluginName)

    @inlineCallbacks
    def stopStandalonePlugin(self, pluginName: str):
        yield self._tryStop(pluginName)

    @inlineCallbacks
    def unloadStandalonePlugin(self, pluginName: str):
        yield self.unloadPlugin(pluginName)

    # ---------------
    # Util methods Plugins

    @inlineCallbacks
    def _tryStart(self, pluginName):
        plugin = self._loadedPlugins[pluginName]
        try:
            yield plugin.start()

        except Exception as e:
            logger.error(
                "An exception occurred while starting plugin %s,"
                " starting continues" % pluginName
            )
            logger.exception(e)

    @inlineCallbacks
    def _tryStop(self, pluginName):
        plugin = self._loadedPlugins[pluginName]
        try:
            yield plugin.stop()

        except Exception as e:
            logger.error(
                "An exception occurred while stopping plugin %s,"
                " stopping continues" % pluginName
            )
            logger.exception(e)

    @inlineCallbacks
    def _unloadPluginPackage(self, pluginName):
        oldLoadedPlugin = self._loadedPlugins.get(pluginName)

        # Stop and remove the Plugin
        del self._loadedPlugins[pluginName]

        try:
            yield oldLoadedPlugin.unload()

        except Exception as e:
            logger.error(
                "An exception occured while unloading plugin %s,"
                " unloading continues" % pluginName
            )
            logger.exception(e)

        # Unload the packages
        loadedSubmodules = [
            modName
            for modName in list(sys.modules.keys())
            if modName.startswith("%s." % pluginName)
        ]

        for modName in loadedSubmodules:
            del sys.modules[modName]

        if pluginName in sys.modules:
            del sys.modules[pluginName]

        gc.collect()

        # pypy doesn't have getrefcount
        # ("oldLoadedPlugin" in this method and the call to getrefcount) == 2 references
        if hasattr(sys, "getrefcount") and sys.getrefcount(oldLoadedPlugin) > 2:
            logger.warning(
                "Old references to %s still exist, count = %s",
                pluginName,
                sys.getrefcount(oldLoadedPlugin),
            )

        del oldLoadedPlugin
        gc.collect()
        # Now there should be no references.

    def sanityCheckServerPlugin(self, pluginName):
        """Sanity Check Plugin

        This method ensures that all the things registed for this plugin are
        prefixed by it's pluginName, EG plugin_noop
        """

        # All endpoint filters must have the 'plugin' : 'plugin_name' in them
        for endpoint in self._vortexEndpointInstancesByPluginName[pluginName]:
            filt = endpoint.filt
            if "plugin" not in filt and filt["plugin"] != pluginName:
                raise Exception(
                    "Payload endpoint does not contain 'plugin':'%s'\n%s"
                    % (pluginName, filt)
                )

        # all tuple names must start with their pluginName
        for tupleName in self._vortexTupleNamesByPluginName[pluginName]:
            TupleCls = tupleForTupleName(tupleName)
            if not tupleName.startswith(
                pluginName
            ) and not tupleName.startswith("vortex._"):
                raise Exception(
                    "Tuple name does not start with '%s', %s (%s)"
                    % (pluginName, tupleName, TupleCls.__name__)
                )

    def notifyOfPluginVersionUpdate(self, pluginName, pluginVersion):
        logger.info(
            "Received PLUGIN update for %s version %s",
            pluginName,
            pluginVersion,
        )
        return self.loadPlugin(pluginName)
